"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.round = exports.RegisterTraceClass = exports.RegisterTraceSetter = exports.RegisterTraceGetter = exports.RegisterTraceValue = exports.isPromise = exports.traceManager = exports.TraceManager = exports.TraceStack = exports.TraceMethod = exports.TraceMethodType = void 0;
const errors_1 = require("./errors");
var TraceMethodType;
(function (TraceMethodType) {
    TraceMethodType[TraceMethodType["Descriptor"] = 0] = "Descriptor";
    TraceMethodType[TraceMethodType["Getter"] = 1] = "Getter";
    TraceMethodType[TraceMethodType["Setter"] = 2] = "Setter";
    TraceMethodType[TraceMethodType["Function"] = 3] = "Function";
})(TraceMethodType = exports.TraceMethodType || (exports.TraceMethodType = {}));
class TraceMethod {
    constructor(className, methodName, methodType) {
        this.className = "";
        this.methodName = "";
        this.source = "";
        this.calls = 0;
        this.inclusiveTime = 0;
        this.innerTime = 0;
        this.methodsCalled = [];
        this.className = className;
        this.methodName = methodName;
        this.calls = 0;
        this.methodType = methodType;
    }
}
exports.TraceMethod = TraceMethod;
class TraceStack {
    constructor(method, args) {
        this.ret = null;
        // timing
        this.completed = false;
        this.startTime = 0;
        this.endTime = 0;
        // stack
        this.stack = "";
        this.level = 0;
        this.method = method;
        this.method.calls++;
        this.callIndex = method.calls;
        this.args = args;
    }
    get systemTime() {
        return new Date().getTime();
    }
    get traceTime() {
        return this.endTime - this.startTime;
    }
    start() {
        this.startTime = this.systemTime;
        // get source line
        // todo: not working properly because of Stub (trace line offset 7 is not correct - it changes)
        this.stack = Error().stack;
        const trace = this.stack.split("\n");
        if (trace.length > 7) {
            const traceSourceInfo = trace[7];
            if (traceSourceInfo.indexOf("(") > 0) {
                const traceLine = /[^(]*\(([^)]*)\)/.exec(traceSourceInfo);
                if (traceLine && traceLine.length > 1) {
                    this.method.source = traceLine[1];
                }
            }
            else {
                const traceLine = /[^\/]*(.*)/.exec(traceSourceInfo);
                if (traceLine && traceLine.length > 1) {
                    this.method.source = traceLine[1];
                }
            }
        }
    }
    complete(ret) {
        this.endTime = this.systemTime;
        this.method.inclusiveTime += this.traceTime;
        this.completed = true;
        this.ret = ret;
    }
    completeException(error) {
        this.complete(undefined);
        this.error = error;
    }
    get argsToString() {
        return this.args.map((val) => `${val}`).join(",");
    }
    toString(showIndent = false, showSource = true, showTiming = false) {
        // todo: error
        // todo: display objects (json!)
        const args = this.argsToString;
        var status = "";
        if (this.completed) {
            if (this.error) {
                status = "‼";
            }
            else {
                status = "◼";
            }
        }
        else {
            status = "▶";
        }
        const source = showSource ? `[${this.method.source}] ` : ``;
        const indent = "".padEnd(showIndent ? this.level * 2 : 0);
        const timing = showTiming ? `${this.traceTime}ms ` : ``;
        if (this.ret) {
            return `${status} ${indent}${this.method.className}.${this.method.methodName}(${args})=${this.ret} ${source}${timing}`;
        }
        else {
            return `${status} ${indent}${this.method.className}.${this.method.methodName}(${args}) ${source}${timing}`;
        }
    }
    toShortString() {
        const args = this.argsToString;
        if (this.ret) {
            return `${this.method.className}.${this.method.methodName}(${args})=${this.ret}`;
        }
        else {
            return `${this.method.className}.${this.method.methodName}(${args})`;
        }
    }
}
exports.TraceStack = TraceStack;
class TraceManager {
    constructor() {
        this.displayTrace = false;
        this.methods = [];
        this.stack = [];
        this.trace = [];
    }
    createMethod(className, methodName, methodType) {
        let method = this.methods.find(x => x.className == className && x.methodName == methodName);
        if (method)
            return method;
        method = new TraceMethod(className, methodName, methodType);
        this.methods.push(method);
        return method;
    }
    start(className, methodName, args, methodType) {
        const method = this.createMethod(className, methodName, methodType);
        const trace = new TraceStack(method, args);
        // add methods called
        const top = this.stack[this.stack.length - 1];
        if (top) {
            if (!top.method.methodsCalled.find(x => x.className == className && x.methodName == methodName)) {
                top.method.methodsCalled.push(method);
            }
        }
        trace.start();
        trace.level = this.stack.length;
        this.stack.push(trace);
        this.trace.push(trace);
        if (this.displayTrace) {
            console.log(`trace| ${trace.toString(true, false, false)}`);
        }
        return trace;
    }
    complete(ret) {
        const last = this.stack.pop();
        if (!last)
            return;
        last.complete(ret);
        // update inner time
        const top = this.stack[this.stack.length - 1];
        if (top) {
            top.method.innerTime += last.traceTime;
        }
    }
    completeError(error) {
        var _a;
        (_a = this.stack.pop()) === null || _a === void 0 ? void 0 : _a.completeException(error);
    }
    displayError(error) {
        console.log(`EXCEPTION name='${error.name}' message='${error.message}'`);
        if (error.name === `mccError` && error.message === `OutsideError`) {
            const innerError = error.innerError;
            if (innerError) {
                console.log(`INNER EXCEPTION name='${innerError.name}' message='${innerError.message}'`);
            }
        }
        console.log(`NODE STACK ${error.stack}`);
        this.showStack();
        this.showTrace(true, false);
    }
    showStack() {
        console.log("TRACE STACK");
        for (let trace of this.stack) {
            console.log(trace.toString());
        }
    }
    showTrace(indent = false, source = true, timing = false) {
        console.log("TRACE");
        for (let trace of this.trace) {
            console.log(trace.toString(indent, source, timing));
        }
    }
    showMethods(indent = false, source = true, timing = false) {
        console.log("METHODS");
        for (let method of this.methods) {
            const exclusiveTime = Math.max(0, method.inclusiveTime - method.innerTime);
            console.log(`◼ ${method.className}.${method.methodName}  ${method.calls} ${method.inclusiveTime}ms ${exclusiveTime}ms`);
            for (let inner of method.methodsCalled) {
                console.log(`    ${inner.className}.${inner.methodName}`);
            }
        }
    }
    clearTrace() {
        this.trace = [];
    }
    get firstTrace() {
        if (this.trace.length === 0)
            return "";
        return this.trace[0].toShortString();
    }
    get lastTrace() {
        if (this.trace.length === 0)
            return "";
        return this.trace[this.trace.length - 1].toShortString();
    }
}
exports.TraceManager = TraceManager;
exports.traceManager = new TraceManager();
function Stub(target, name, funct, cx, args, methodType) {
    exports.traceManager.start(target.name, name, args, methodType);
    try {
        let res = methodType === TraceMethodType.Getter ? funct.apply(cx) : funct.apply(cx, args);
        if (!isPromise(res)) {
            exports.traceManager.complete(res);
            return res;
        }
        return new Promise((resolve, reject) => {
            res.then((result) => {
                exports.traceManager.complete(result);
                resolve(result);
            }).catch((error) => {
                if ((error === null || error === void 0 ? void 0 : error.name) === errors_1.MCC_ERROR) {
                    exports.traceManager.displayError(error);
                    exports.traceManager.completeError(error);
                    reject(error);
                }
                else {
                    const newError = new errors_1.mccOutsideError(error);
                    exports.traceManager.displayError(newError);
                    exports.traceManager.completeError(newError);
                    reject(newError);
                }
            });
        });
    }
    catch (error) {
        if ((error === null || error === void 0 ? void 0 : error.name) === errors_1.MCC_ERROR) {
            exports.traceManager.displayError(error);
            exports.traceManager.completeError(error);
            throw error;
        }
        else {
            const newError = new errors_1.mccOutsideError(error);
            exports.traceManager.displayError(newError);
            exports.traceManager.completeError(newError);
            throw new errors_1.mccOutsideError(newError);
        }
    }
}
function isPromise(p) {
    if (typeof p === 'object' && typeof p.then === 'function') {
        return true;
    }
    return false;
}
exports.isPromise = isPromise;
function RegisterTraceValue(target, name, descriptor) {
    //  decorating a method
    if (descriptor && descriptor.value) {
        let original = descriptor.value;
        descriptor.value = function (...args) {
            return Stub(target, name, original, this, args, TraceMethodType.Descriptor);
        };
        return descriptor;
    }
    return null;
}
exports.RegisterTraceValue = RegisterTraceValue;
function RegisterTraceGetter(target, name, descriptor) {
    if (descriptor && descriptor.get) {
        let original = descriptor.get;
        descriptor.get = function () {
            return Stub(target, name, original, this, [], TraceMethodType.Getter);
        };
        return descriptor;
    }
    return null;
}
exports.RegisterTraceGetter = RegisterTraceGetter;
function RegisterTraceSetter(target, name, descriptor) {
    if (descriptor && descriptor.set) {
        let original = descriptor.set;
        descriptor.set = function (...args) {
            return Stub(target, name, original, this, args, TraceMethodType.Setter);
        };
        return descriptor;
    }
    return null;
}
exports.RegisterTraceSetter = RegisterTraceSetter;
function RegisterTraceClass(targetClass) {
    // decorating a class
    // add tracing capability to all own methods (doesn't work for constructor)
    Object.getOwnPropertyNames(targetClass.prototype).forEach((methodName) => {
        const desc = Object.getOwnPropertyDescriptor(targetClass.prototype, methodName);
        // getter
        if (desc === null || desc === void 0 ? void 0 : desc.get) {
            RegisterTraceGetter(targetClass, methodName, desc);
            Object.defineProperty(targetClass.prototype, methodName, desc);
            return;
        }
        // setter
        if (desc === null || desc === void 0 ? void 0 : desc.set) {
            RegisterTraceSetter(targetClass, methodName, desc);
            Object.defineProperty(targetClass.prototype, methodName, desc);
            return;
        }
        let original = targetClass.prototype[methodName];
        // constructor
        if (methodName === "constructor") {
            RegisterTraceValue(targetClass, methodName, desc);
            Object.defineProperty(targetClass.prototype, methodName, desc);
            return;
        }
        // skip non functions
        if (typeof original !== "function") {
            return;
        }
        // an arrow function can't be used while we have to preserve right 'this'
        targetClass.prototype[methodName] = function (...args) {
            return Stub(targetClass, methodName, original, this, args, TraceMethodType.Function);
        };
    });
    return targetClass;
}
exports.RegisterTraceClass = RegisterTraceClass;
function round(x, decimal = 0) {
    if (decimal === 0)
        return Math.round(x);
    const dec10 = Math.pow(10, decimal);
    return Math.round(x * dec10) / dec10;
}
exports.round = round;
//# sourceMappingURL=trace.js.map